/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sftp.connection;

import org.ow2.petals.bc.sftp.ConfigurationException;
import org.ow2.petals.bc.sftp.SFTPConstants;
import org.ow2.petals.bc.sftp.authentication.SshAuthentication;
import org.ow2.petals.bc.sftp.authentication.SshPasswordAuthentication;
import org.ow2.petals.bc.sftp.authentication.SshPublicKeyAuthentication;
import org.ow2.petals.bc.sftp.authentication.exception.AuthenticationException;
import org.ow2.petals.bc.sftp.connection.exception.ConnectionException;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.jbidescriptor.generated.Retrypolicy;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.xml.XMLHelper;

/**
 * Helper class which helps (sic) building SFTPConnectionInfo instance using the
 * values given in the SU ConfigurationExtensions or using the configuration
 * node given in a JBI message.
 * 
 * @author Charles Casadei - EBM WebSourcing
 */
public class SFTPConnectionInfoBuilder {

    /**
     * Builds a new instance of SFTPConnection info using the informations from
     * the given node. This allow the SFTP component to be parameterized at
     * runtime.
     * 
     * @param node
     *            the service unit configuration
     * @return a new SFTPConnectionInfo instance
     * @throws ConfigurationException
     *             if the service configuration is not valid
     */
    public static SFTPConnectionInfo buildSFTPConnectionInfo(Node node)
            throws ConfigurationException {

        Node serverNode = null;
        Node portNode = null;
        Node userNode = null;
        Node passwordNode = null;
        Node privateKeyNode = null;
        Node passphraseNode = null;
        Node directoryNode = null;
        Node overwrite = null;

        SshAuthentication sshAuthentication = null;

        // Retrieve the configuration nodes from the given "root" node
        serverNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_SERVER, false);
        portNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_PORT, false);
        userNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_USER, false);
        passwordNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_PASSWORD, false);
        privateKeyNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_PRIVATE_KEY, false);
        passphraseNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_PASSPHRASE, false);
        directoryNode = XMLHelper.findChild(node, null, SFTPConstants.SFTP_FOLDER, false);
        overwrite = XMLHelper.findChild(node, null, SFTPConstants.SFTP_OVERWRITE, false);

        // If the user node is empty
        if (userNode == null) {
            throw new ConfigurationException("Missing configuration element : "
                    + SFTPConstants.SFTP_USER);
        }

        // Checks if there's only one authentication mode (ie if password => no
        // private key & no passphrase
        if (((passwordNode != null) && (privateKeyNode != null))
                || ((passwordNode != null) && (passphraseNode != null))) {
            throw new ConfigurationException(
                    "Can't use both password authentication and private key authentication");
        }

        // A passphrase must match a private key...if no private key -> bang :-)
        if ((passphraseNode != null) && (privateKeyNode == null)) {
            throw new ConfigurationException("A passphrase with no matching private key was found");
        }

        // If the authentication method is : user / password
        if (passwordNode != null) {
            try {
                sshAuthentication = new SshPasswordAuthentication(userNode.getTextContent(),
                        passwordNode.getTextContent());
            } catch (final AuthenticationException e) {
                throw new ConfigurationException(e);
            }
        }
        // Else if the authentication method is user / privatekey / passphrase
        else if (privateKeyNode != null) {
            try {
                sshAuthentication = new SshPublicKeyAuthentication(userNode.getTextContent(),
                        privateKeyNode.getTextContent(),
                        ((passphraseNode != null) ? passphraseNode.getTextContent() : null));
            } catch (final AuthenticationException e) {
                throw new ConfigurationException(e);
            }
        } else {
            throw new ConfigurationException("No authentications information found");
        }

        // Builds a SFTPConnectionInfo object from the previous nodes
        try {
            return new SFTPConnectionInfo(((serverNode != null) ? serverNode.getTextContent()
                    : null), ((portNode != null) ? portNode.getTextContent() : null),
                    sshAuthentication, ((directoryNode != null) ? directoryNode.getTextContent()
                            : null), overwrite != null ? Boolean.parseBoolean(overwrite
                            .getTextContent()) : false);
        } catch (final ConnectionException e) {
            throw new ConfigurationException(e);
        }
    }

    /**
     * TODO: move that in a Validator extension Builds a new instance of
     * SFTPConnectionInfo using the information given in the service unit
     * descriptor.
     * 
     * @param extensions
     *            the service unit configuration
     * @return a new SFTPConnectionInfo instance
     * @throws ConfigurationException
     *             if the service configuration is not valid
     */
    public static SFTPConnectionInfo buildSFTPConnectionInfo(
            final ConfigurationExtensions extensions, final Retrypolicy retryPolicy)
            throws ConfigurationException {

        SshAuthentication sshAuthentication = null;

        final String server = extensions.get(SFTPConstants.SFTP_SERVER);
        final String port = extensions.get(SFTPConstants.SFTP_PORT);
        final String user = extensions.get(SFTPConstants.SFTP_USER);
        final String password = extensions.get(SFTPConstants.SFTP_PASSWORD);
        final String privateKey = extensions.get(SFTPConstants.SFTP_PRIVATE_KEY);
        final String passphrase = extensions.get(SFTPConstants.SFTP_PASSPHRASE);
        final String directory = extensions.get(SFTPConstants.SFTP_FOLDER);
        final boolean overwrite = Boolean
                .parseBoolean(extensions.get(SFTPConstants.SFTP_OVERWRITE));
        int maxConnection = SFTPConstants.DEFAULT_MAX_CONNECTION;
        long maxIdleTime = SFTPConstants.DEFAULT_SSH_MAX_IDLE_TIME;

        try {
            maxIdleTime = Long.parseLong(extensions.get(SFTPConstants.SFTP_MAX_IDLE_TIME));
        } catch (NumberFormatException e) {
        }

        try {
            maxConnection = Integer.parseInt(extensions.get(SFTPConstants.SFTP_MAX_CONNECTION));
        } catch (NumberFormatException e) {
        }

        // Checks if there's only one authentication mode (ie if password => no
        // private key & no passphrase
        if (((password != null) && (privateKey != null))
                || ((password != null) && (passphrase != null))) {
            throw new ConfigurationException(
                    "Can't use both password authentication and private key authentication");
        }

        // A passphrase must match a private key...if no private key -> bang :-)
        if ((passphrase != null) && (privateKey == null)) {
            throw new ConfigurationException("A passphrase with no matching private key was found");
        }

        // If the authentication method is : user / password
        if (password != null) {
            try {
                sshAuthentication = new SshPasswordAuthentication(user, password);
            } catch (final AuthenticationException e) {
                throw new ConfigurationException(e);
            }
        }
        // Else if the authentication method is user / privatekey / passphrase
        else if (privateKey != null) {
            try {
                sshAuthentication = new SshPublicKeyAuthentication(user, privateKey, passphrase);
            } catch (final AuthenticationException e) {
                throw new ConfigurationException(e);
            }
        } else {
            throw new ConfigurationException("No authentication informations found");
        }

        try {
            SFTPConnectionInfo sftpConnectionInfo = new SFTPConnectionInfo(server, port,
                    sshAuthentication, directory, overwrite);
            sftpConnectionInfo.setMaxConnection(maxConnection);
            sftpConnectionInfo.setMaxIdleTime(maxIdleTime);
            return sftpConnectionInfo;
        } catch (final ConnectionException e) {
            throw new ConfigurationException(e);
        }
    }
}
