/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sftp.connection;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.pool.PoolableObjectFactory;

import com.sshtools.j2ssh.SftpClient;
import com.sshtools.j2ssh.SshEventAdapter;
import com.sshtools.j2ssh.transport.TransportProtocol;

/**
 * This factory creates instances of WrappedSftpClient. 
 * <p>
 * It implements the lifecycle of the pool.
 * </p>
 * 
 * @author Mathieu CARROLLE - EBM WebSourcing
 */
public class WrappedSftpClientFactory implements PoolableObjectFactory {

    /**
     * This class disconnect client (ssh and sftp) when a socket timeout occurs.
     * 
     * @author ”Mathieu CARROLLE - mathieu.carrolle@petalslink.com”
     * 
     */
    private class EventHandler extends SshEventAdapter {
        final WrappedSftpClient client;

        public EventHandler(WrappedSftpClient client) {
            this.client = client;
        }

        @Override
        public void onConnected(TransportProtocol arg0) {
        }

        @Override
        public void onDisconnect(TransportProtocol arg0) {
        }

        /**
         * Remove all pending connection
         */
        @Override
        public void onSocketTimeout(TransportProtocol arg0) {
            logger.finest("Socket timeout occurs [" + arg0.getUnderlyingProviderDetail() + "]");
            try {
                destroyObject(this.client);
            } catch (Exception e) {
            }
        }
    }

    private final SFTPConnectionInfo connectionConfiguration;

    private final Logger logger;

    public WrappedSftpClientFactory(final SFTPConnectionInfo connectionConfiguration,
            final Logger logger) {
        this.connectionConfiguration = connectionConfiguration;
        this.logger = logger;
    }

    @Override
    public void activateObject(Object obj) throws Exception {
    }

    @Override
    public void destroyObject(Object obj) throws Exception {
        WrappedSftpClient client = (WrappedSftpClient) obj;
        if (client != null) {
            SftpClient activeSftpClient = client.getActiveSftpClient();
            if (this.logger.isLoggable(Level.FINEST))
                this.logger.finest("Log out user and disconnect ssh session : ["
                        + connectionConfiguration.toString() + "]");
            if (activeSftpClient != null && !activeSftpClient.isClosed()) {
                activeSftpClient.quit();
            }
            if (client.isConnected()) {
                client.disconnect();
            }
            client = null;
        }
    }

    @Override
    public Object makeObject() throws Exception {
        this.logger.finest("Create new instance of SFTPClient");
        final WrappedSftpClient client = new WrappedSftpClient(connectionConfiguration, logger);
        client.addEventHandler(new EventHandler(client));
        client.connectAndLog();
        client.configure();
        return client;
    }

    @Override
    public void passivateObject(Object obj) throws Exception {
    }

    @Override
    public boolean validateObject(Object obj) {
        this.logger.finest("Validate instance of SFTPClient");
        final WrappedSftpClient client = (WrappedSftpClient) obj;
        SftpClient activeSftpClient;
        if (client == null) {
            return false;
        }
        try {
            activeSftpClient = client.getActiveSftpClient();
            if (activeSftpClient != null && !activeSftpClient.isClosed() && client.isConnected()) {
                this.logger.finest("SFTPClient is connected");
                return true;
            }
        } catch (IOException e) {
        }
        return false;
    }
}
