/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sftp.listeners;

import java.io.IOException;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.bc.sftp.ConfigurationException;
import org.ow2.petals.bc.sftp.SFTPComponent;
import org.ow2.petals.bc.sftp.SFTPConstants;
import org.ow2.petals.bc.sftp.SFTPUtil;
import org.ow2.petals.bc.sftp.connection.SFTPConnectionInfo;
import org.ow2.petals.bc.sftp.connection.SFTPConnectionInfoBuilder;
import org.ow2.petals.bc.sftp.connection.WrappedSftpClient;
import org.ow2.petals.bc.sftp.service.SFTPService;
import org.ow2.petals.component.framework.api.Wsdl;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.WSDLUtilImpl;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;
import com.sshtools.j2ssh.SftpClient;

/**
 * The SFTP JBI listener writes a file to SFTP server on incoming JBI message.
 * @author Charles Casadei - EBM WebSourcing
 */
public class JBIListener extends AbstractJBIListener {

    /**
     * A reference on the effective SFTP service.
     */
    private transient SFTPService service = null;

    /**
     * Validate the operation.
     * 
     * @param operation
     * @throws MessagingException
     */
    private static final void validateExchange(final Exchange exchange) throws MessagingException {
        final QName operation = exchange.getOperation();
        // Null operation
        if (operation == null) {
            throw new MessagingException(
                    "The JBI operation must be set by the consumer of the service.");
        }
        // if no namespace is provided we use the default namespace.
        if (StringHelper.isNullOrEmpty(operation.getNamespaceURI())) {
            exchange.setOperation(new QName(SFTPConstants.SFTP_SERVICE_NS, operation.getLocalPart()));
        }
    }

    /**
     * Initialize the JBI Listener.
     */
    @Override
    public void init() {
        service = new SFTPService(getLogger());
    }

    /**
     * Process the exchange depending on the operation : <li><b>dir</b> return
     * the filenames list or the remote directory</li> <li><b>get</b> return the
     * remote file with the name specified in the exchange (as attachment or
     * source, depending on the SU property)</li> <li><b>mget</b> return all the
     * remote files</li> <li><b>put</b> put the message</li> <li><b>mput</b> put
     * attachments</li>
     * 
     * @param exchange
     *            message received by the component
     */
    @Override
    public boolean onJBIMessage(Exchange exchange) {
        if (exchange.isActiveStatus()) {

            GenericObjectPool pool = null;
            WrappedSftpClient client = null;
            try {
                validateExchange(exchange);
                final String operationName = exchange.getOperationName();
                final Document inDocument = exchange.getInMessageContentAsDocument();
                final Wsdl nativeWSDL = this.component.getNativeWsdl();

                // test if the component exposes a generic service
                if (nativeWSDL != null
                        && WSDLUtilImpl.isDescriptionContaining(nativeWSDL.getDescription(),
                                exchange.getEndpoint())) {
                    // search a "connection" elements in the children of the
                    // document
                    final Node connectionNode = XMLHelper.findChild(inDocument.getFirstChild(),
                            null, SFTPConstants.SFTP_CONNECTION, false);
                    final SFTPConnectionInfo connectionInfo = SFTPConnectionInfoBuilder
                            .buildSFTPConnectionInfo(connectionNode);
                    client = new WrappedSftpClient(connectionInfo, this.getLogger());
                    client.connectAndLog();
                    client.configure();
                } else {
                    // Not generic service call
                    pool = ((SFTPComponent) this.component).getMapOfPool().get(this.getProvides());
                    if (this.getLogger().isLoggable(Level.FINEST)) {
                        this.getLogger().finest(
                                "Pool status [idleInstance, activeInstance]: [" + pool.getNumIdle()
                                        + "," + pool.getNumActive() + "]");
                    }
                    client = (WrappedSftpClient) pool.borrowObject();
                }

                if (SFTPConstants.DIR.equalsIgnoreCase(operationName)) {
                    service.processDIR(exchange, inDocument, client);
                } else if (SFTPConstants.PUT.equalsIgnoreCase(operationName)) {
                    service.processPUT(exchange, inDocument, client);
                } else if (SFTPConstants.MPUT.equalsIgnoreCase(operationName)) {
                    service.processMPUT(exchange, inDocument, client);
                } else if (SFTPConstants.GET.equalsIgnoreCase(operationName)) {
                    service.processGET(exchange, inDocument, client);
                } else if (SFTPConstants.GETASATTACHMENT.equalsIgnoreCase(operationName)) {
                    service.processGetAsAttachment(exchange, inDocument, client);
                } else if (SFTPConstants.MGET.equalsIgnoreCase(operationName)) {
                    service.processMGET(exchange, inDocument, client);
                } else if (SFTPConstants.DEL.equalsIgnoreCase(operationName)) {
                    service.processDel(exchange, inDocument, client);
                } else {
                    final String message = "The operation '" + exchange.getOperation()
                            + "' is not recognized";
                    throw new MessagingException(message);
                }
            } catch (ConfigurationException e) {
                if (!exchange.isInOnlyPattern()) {
                    try {
                        SFTPUtil.setConfigurationFaultOnExchange(exchange, e.getMessage());
                    } catch (MessagingException e1) {
                        exchange.setError(e1);
                    }
                } else {
                    exchange.setError(e);
                }
            } catch (IOException e) {
                if (!exchange.isInOnlyPattern()) {
                    try {
                        SFTPUtil.setIOFaultOnExchange(exchange, e.getMessage());
                    } catch (MessagingException e1) {
                        exchange.setError(e1);
                    }
                } else {
                    exchange.setError(e);
                }
            } catch (Exception e) {
                this.getLogger().log(Level.WARNING,
                        e.getClass().getSimpleName() + " : " + e.getMessage());
                exchange.setError(e);
            } finally {
                try {
                    if (client != null) {
                        if (pool == null) {
                            this.getLogger().finest(
                                    "Disconnect SFTPClient for the generic service connection");
                            SftpClient activeSftpClient = client.getActiveSftpClient();
                            if (activeSftpClient != null && !activeSftpClient.isClosed()) {
                                activeSftpClient.quit();
                            }
                            if (client.isConnected()) {
                                client.disconnect();
                            }
                            client = null;
                        } else {
                            this.getLogger().finest("Return FTPClient instance to the pool");
                            pool.returnObject(client);
                        }
                    }
                } catch (IOException e) {
                    this.getLogger().finest(
                            "Cannot disconnect the SFTPClient for the generic service connection : ["
                                    + e.getMessage() + "]");
                } catch (Exception e2) {
                    this.getLogger().finest(
                            "Cannot release object of the pool : [" + e2.getMessage() + "]");
                }
            }
        }

        return true;
    }
}
