/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sftp.service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;

import org.ow2.petals.bc.sftp.ConfigurationException;
import org.ow2.petals.bc.sftp.SFTPConstants;
import org.ow2.petals.bc.sftp.SFTPUtil;
import org.ow2.petals.bc.sftp.SftpProvideExtFlowStepBeginLogData;
import org.ow2.petals.bc.sftp.connection.WrappedSftpClient;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.logger.ProvideExtFlowStepEndLogData;
import org.ow2.petals.component.framework.util.MtomUtil;
import org.ow2.petals.component.framework.util.MtomUtil.MtomMapping;
import org.ow2.petals.component.framework.util.SourceUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;

/**
 * The SFTP JBI listener writes a file to SFTP server on incoming JBI message.
 * @author Charles Casadei - EBM WebSourcing
 */
public class SFTPService {

    /**
     * The logger used to log informations relative to this class.
     */
    private final Logger logger;

    /**
     * Default constructor.
     * 
     * @param logger
     *            logger used to log informations relative to this class.
     */
    public SFTPService(final Logger logger) {
        super();
        this.logger = logger;
    }

    /**
     * Lists files of the remote working directory. <br />
     * Supported patterns are : <li>InOut <li>inOptionalOut <br />
     * No attachments are expected <br />
     * No fault / error is set on the message exchange by this method. This
     * method wraps the errors into a MessagingException which must be handled
     * at higher level.
     * 
     * @param exchange
     *            the incoming message exchange
     * @param connectionInfo
     *            information on the SFTP server
     * @throws MessagingException
     *             if any problem is spotted during execution
     */
    public void processDIR(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException {
        this.logger.info("JBI exchange receive - Process DIR operation");

        // Checks the MEP of the incoming message
        if (!(exchange.isInOutPattern())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.DIR,
                    Message.MEPConstants.IN_OUT_PATTERN));
            throw new MessagingException(errorSB.toString());
        }

        // Check if the message has attachements
        if (!exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("No attachement expected.");
            throw new MessagingException(errorSB.toString());
        }

        // Checking the message content
        final Element root = inDocument.getDocumentElement();
        // If the first element of the message is not the right one
        if (root == null || !SFTPConstants.DIR.equalsIgnoreCase(root.getLocalName())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Invalid message content for the DIR operation.");
            throw new MessagingException(errorSB.toString());
        }

        // Performs the DIR operation
        List<String> fileNames = null;
        try {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Retrieve the list of files
            fileNames = client.list();

            ProvideExtFlowStepEndLogData sftpProvideExtFlowStepEndLogData = new ProvideExtFlowStepEndLogData(
                    flowInstanceId, flowStepId);
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepEndLogData);
            
            // generate filename list
            final String result = SFTPUtil.generateFileNameList(fileNames, exchange.getOperation());
            // Sets the exchange OUT content
            exchange.getOutMessage().setContent(SourceUtil.createSource(result));
        } catch (final NullPointerException ex) {
            final String error = "Can not retrieve file list from the SFTP server ["
                    + client.getInfo() + "]. Cause : Socket timeout occurs";
            throw new MessagingException(error);
        } catch (final IOException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not retrieve file list from the SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. IOException : ");
            errorSB.append(ex.getMessage());
            SFTPUtil.setIOFaultOnExchange(exchange, errorSB.toString());
        } catch (final PEtALSCDKException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not retrieve file list from the SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. PetalsCDKException : ");
            errorSB.append(ex.getMessage());
            SFTPUtil.setIOFaultOnExchange(exchange, errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("DIR operation ended");
        }
    }

    /**
     * Puts the content on the incoming XML message into the specified file, on
     * the SFTP server. <br />
     * Supported patterns are : <li>InOnly <li>inOptionalOut <li>RobustInOnly <br />
     * No attachments are expected <br />
     * No fault / error is set on the message exchange by this method. This
     * method wraps the errors into a MessagingException which must be handled
     * at higher level.
     * 
     * @param exchange
     *            the incoming message exchange
     * @param connectionInfo
     *            information on the SFTP server
     * @throws MessagingException
     *             if any problem is spotted during execution
     */
    public void processPUT(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException {
        this.logger.info("JBI exchange receive - Process PUT operation");
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.PUT,
                    Message.MEPConstants.IN_ONLY_PATTERN));
            throw new MessagingException(errorSB.toString());
        }

        // Check if the message has attachements
        if (!exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("No attachement expected.");
            throw new MessagingException(errorSB.toString());
        }

        // Checking the message content
        final Node root = inDocument.getFirstChild();
        // If the first element of the message is not the right one
        if (root == null || !SFTPConstants.PUT.equalsIgnoreCase(root.getLocalName())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Invalid message content for the PUT operation");
            throw new MessagingException(errorSB.toString());
        }

        final List<String> filenames = new ArrayList<String>();
        // Retrieve all the node under the root node
        final NodeList rootChildNodes = root.getChildNodes();
        // Iterate through the childs of the root node
        for (int i = 0; i < rootChildNodes.getLength(); i++) {
            final Node node = rootChildNodes.item(i);
            if (node != null && SFTPConstants.SFTP_FILENAME.equalsIgnoreCase(node.getLocalName())) {
                filenames.add(node.getTextContent());
            }
        }

        // Check if there is at least ONE filename in the list
        if (filenames.isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't retrieve the filename from the XML content : ");
            errorSB.append("No filename node found.");
            throw new MessagingException(errorSB.toString());
        }

        // Checks if there is MORE THAN ONE filenames
        if (filenames.size() > 1) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't retrieve the filename from the XML content : ");
            errorSB.append("Too many filenames nodes found.");
            throw new MessagingException(errorSB.toString());
        }

        // Checks the filename validity
        final String fileName = filenames.get(0);
        // If the filename is null of ""
        if (StringHelper.isNullOrEmpty(fileName)) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can't retrieve the filename from the XML content : ");
            errorSB.append("missing filename.");
            throw new MessagingException(errorSB.toString());
        }

        // Check if there is a content node
        final Node bodyNode = XMLHelper.findChild(root, null, SFTPConstants.BODY, false);
        // If the content node can not be found
        if (bodyNode == null || !SFTPConstants.BODY.equalsIgnoreCase(bodyNode.getLocalName())) {
            this.logger.warning("Can't retrieve message body.");
        }

        // Performs the PUT operation
        try {
            String content = XMLHelper.toString(bodyNode.getChildNodes());
            
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }

            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, fileName, client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Put the content to the SSH server
            client.putString(fileName, content);
        } catch (final IOException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. IOException : ");
            errorSB.append(ex.getMessage());
            throw new MessagingException(errorSB.toString());
        } catch (final NullPointerException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put XML content on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. Cause : Socket timeout occurs");
            throw new MessagingException(errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("PUT operation ended");
        }
    }

    public void processMPUT(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process MPUT operation");
        }
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            throw new MessagingException(SFTPUtil.getValidMepMessageForOperation(
                    SFTPConstants.MPUT, MEPConstants.IN_ONLY_PATTERN));
        }
        // Check if the message has attachements
        if (exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("One or more attachments expected.");
            throw new MessagingException(errorSB.toString());
        }
        // Perform the MPUT operation
        try {
            MtomMapping mapping = MtomUtil
                    .getMtomMapping(exchange, inDocument.getDocumentElement());
            Map<String, DataHandler> attachments = mapping.getContentIdToDataHandler();

            // TODO to factorize
            StringBuffer filePattern =new StringBuffer("(");
            int fileNamesLength = attachments.size();
            Iterator<String> attachmentsKeyIterator = attachments.keySet().iterator();
            int i = 0;
            while(attachmentsKeyIterator.hasNext()) {
                filePattern.append(attachmentsKeyIterator.next());
                if(i <  fileNamesLength - 1) {
                    filePattern.append("|");
                }
                i++;
            }
            filePattern.append(")");
            
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }
            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, filePattern.toString(), client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Sends the attachments to the SFTP server
            client.mput(attachments);
        } catch (final NullPointerException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put file on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. Cause : Socket timeout occurs");
            throw new MessagingException(errorSB.toString());
        } catch (final IOException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not put file on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. IOException : ");
            errorSB.append(ex.getMessage());
            throw new MessagingException(errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("MPUT operation ended");
        }
    }

    /**
     * Gets the specified file from the SFTP server. <br />
     * Supported pattern are : <li>InOut <li>
     * No attachments are expected <br />
     * No fault / error is set on the message exchange by this method. This
     * method wraps the errors into a MessagingException which must be handled
     * at higher level.
     * 
     * @param exchange
     *            the incoming message exchange
     * @param connectionInfo
     *            information on the SFTP server
     * @throws MessagingException
     *             if any problem is spotted during execution
     */
    public void processGET(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException, ConfigurationException {
        this.logger.info("JBI exchange receive - Process GET operation");

        // Checks the MEP of the incoming message
        if (!(exchange.isInOutPattern())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.GET,
                    Message.MEPConstants.IN_OUT_PATTERN));
            throw new MessagingException(errorSB.toString());
        }

        // Check if the message has attachements
        if (!exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("No attachement expected.");
            throw new MessagingException(errorSB.toString());
        }
        // Check the root node
        final Node root = inDocument.getFirstChild();
        if (root == null || !SFTPConstants.GET.equalsIgnoreCase(root.getLocalName())) {
            throw new ConfigurationException("Missing root element : " + SFTPConstants.GET);
        }
        // retrieve the filename to get
        final Node filePatternNode = XMLHelper.findChild(root, null, SFTPConstants.SFTP_FILENAME,
                true);
        String filePattern = null;
        if (filePatternNode != null) {
            filePattern = filePatternNode.getTextContent();
        }
        if (StringHelper.isNullOrEmpty(filePattern))
            throw new ConfigurationException("Missing request element :"
                    + SFTPConstants.SFTP_FILENAME);

        // Perform the GET operation
        try {

            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }
            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, filePattern, client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Gets the file from the SFTP server
            final Document result = client.get(filePattern);
            
            ProvideExtFlowStepEndLogData sftpProvideExtFlowStepEndLogData = new ProvideExtFlowStepEndLogData(
                    flowInstanceId, flowStepId);
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepEndLogData);
            
            SFTPUtil.addRootElement(result, new QName(exchange.getOperation().getNamespaceURI(),
                    "ver:getResponse"));
            // Put the file in the message
            exchange.setOutMessageContent(result);
        } catch (final NullPointerException ex) {
            final String error = "Can not get file [" + filePattern + "] on SFTP server ["
                    + client.getInfo().getServer() + "]. Cause : Socket timeout occurs";
            throw new MessagingException(error);
        } catch (final IOException ex) {
            SFTPUtil.setIOFaultOnExchange(exchange,
                    "Can not get file [" + filePattern + "] on SFTP server ["
                            + client.getInfo().getServer() + "]. Cause : " + ex.getMessage());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("GET operation ended");
        }
    }

    /**
     * Gets the specified file from the SFTP server. <br />
     * Supported pattern are : <li>InOut <li>
     * The specified file is returned as out-messsage attachment.
     * 
     * @param exchange
     *            the incoming message exchange
     * @param connectionInfo
     *            information on the SFTP server
     * @throws MessagingException
     *             if any problem is spotted during execution
     */
    public void processGetAsAttachment(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException, ConfigurationException {
        this.logger.info("JBI exchange receive - Process GET operation");

        // Checks the MEP of the incoming message
        if (!(exchange.isInOutPattern())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.GETASATTACHMENT,
                    Message.MEPConstants.IN_OUT_PATTERN));
            throw new MessagingException(errorSB.toString());
        }

        // Check if the message has attachements
        if (!exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("No attachement expected.");
            throw new MessagingException(errorSB.toString());
        }
        // Check the root node
        final Node root = inDocument.getFirstChild();
        if (root == null || !SFTPConstants.GETASATTACHMENT.equalsIgnoreCase(root.getLocalName())) {
            throw new ConfigurationException("Missing root element : "
                    + SFTPConstants.GETASATTACHMENT);
        }
        // retrieve the filename to get
        final Node filePatternNode = XMLHelper.findChild(root, null, SFTPConstants.SFTP_FILENAME,
                true);
        String filePattern = null;
        if (filePatternNode != null) {
            filePattern = filePatternNode.getTextContent();
        }
        if (StringHelper.isNullOrEmpty(filePattern))
            throw new ConfigurationException("Missing request element : "
                    + SFTPConstants.SFTP_FILENAME);
        // Perform the GET operation
        try {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }
            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, filePattern, client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Gets the file from the SFTP server
            final DataHandler retrievedFile = client.getAsAttachment(filePattern);
            
            ProvideExtFlowStepEndLogData sftpProvideExtFlowStepEndLogData = new ProvideExtFlowStepEndLogData(
                    flowInstanceId, flowStepId);
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepEndLogData);
            
            final Document response = SFTPUtil.generateMTOMResponse(retrievedFile.getName(),
                    exchange.getOperation());

            exchange.setOutMessageAttachment(retrievedFile.getName(), retrievedFile);
            exchange.setOutMessageContent(response);
        } catch (final NullPointerException ex) {
            final String error = "Can not get file [" + filePattern + "] on SFTP server ["
                    + client.getInfo().getServer() + "]. Cause : Socket timeout occurs";
            throw new MessagingException(error);
        } catch (final IOException ex) {
            SFTPUtil.setIOFaultOnExchange(exchange,
                    "Can not get file [" + filePattern + "] on SFTP server ["
                            + client.getInfo().getServer() + "]. Cause : " + ex.getMessage());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("GET operation ended");
        }
    }

    /**
     * Gets the specified files from the SFTP server. <br />
     * Supported patterns are : <li>InOut <li>inOptionalOut <br />
     * No attachments are expected <br />
     * No fault / error is set on the message exchange by this method. This
     * method wraps the errors into a MessagingException which must be handled
     * at higher level.
     * 
     * @param exchange
     *            the incoming message exchange
     * @param connectionInfo
     *            information on the SFTP server
     * @throws MessagingException
     *             if any problem is spotted during execution
     */
    public void processMGET(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException, ConfigurationException {
        this.logger.info("JBI exchange receive - Process MGET operation");

        // Checks the MEP of the incoming message
        if (!(exchange.isInOutPattern())) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.MGET,
                    Message.MEPConstants.IN_OUT_PATTERN));
            throw new MessagingException(errorSB.toString());
        }

        // Check if the message has attachements
        if (!exchange.getInMessageAttachments().isEmpty()) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("No attachement expected.");
            throw new MessagingException(errorSB.toString());
        }

        // Check the root node
        final Node root = inDocument.getFirstChild();
        if (root == null || !SFTPConstants.MGET.equalsIgnoreCase(root.getLocalName())) {
            throw new ConfigurationException("Missing root element : " + SFTPConstants.MGET);
        }

        // retrieve filenames to get
        final NodeList list = root.getChildNodes();
        final List<String> fileNames = new LinkedList<String>();        
        for (int index = 0; index < list.getLength(); index++) {
            final Node node = list.item(index);
            if (SFTPConstants.SFTP_FILENAME.equalsIgnoreCase(node.getLocalName())) {
                String fileName = node.getTextContent();
                fileNames.add(fileName);
            }
        }
        
        if (fileNames.isEmpty()) {
            throw new ConfigurationException("Missing request element : "
                    + SFTPConstants.SFTP_FILENAME);
        }
        // Perform the MGET operation
        try {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }

            // TODO mutualize code
            StringBuffer filePattern =new StringBuffer("(");
            int fileNamesLength = fileNames.size();
            for(int i = 0; i < fileNamesLength; i++) {
                filePattern.append(fileNames.get(i));
                if(i <  fileNamesLength - 1) {
                    filePattern.append("|");
                }
            }
            filePattern.append(")");
            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, filePattern.toString(), client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            // Perform the MGET operation
            final Map<String, DataHandler> files = client.mGet(fileNames);
            
            ProvideExtFlowStepEndLogData sftpProvideExtFlowStepEndLogData = new ProvideExtFlowStepEndLogData(
                    flowInstanceId, flowStepId);
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepEndLogData);
            
            // Add attachments on the OUT message
            for (final Entry<String, DataHandler> filesEntry : files.entrySet()) {
                exchange.getOutMessage().addAttachment(filesEntry.getKey(), filesEntry.getValue());
            }
            // Builds the list of the retrieved files
            final List<String> retrievedFilenames = new LinkedList<String>(files.keySet());
            // Builds the response content
            final Document outDoc = SFTPUtil.generateMTOMListResponse(retrievedFilenames,
                    exchange.getOperation());
            exchange.setOutMessageContent(outDoc);
        } catch (final NullPointerException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not get files on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. Cause : Socket timeout occurs");
            throw new MessagingException(errorSB.toString());
        } catch (final IOException ex) {
            final StringBuilder errorSB = new StringBuilder();
            errorSB.append("Can not get file on SFTP server [");
            errorSB.append(client.getInfo());
            errorSB.append("]. IOException : ");
            errorSB.append(ex.getMessage());
            SFTPUtil.setIOFaultOnExchange(exchange, errorSB.toString());
        }
        if (this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("MGET operation ended");
        }
    }

    public void processDel(final Exchange exchange, final Document inDocument,
            final WrappedSftpClient client) throws MessagingException, ConfigurationException {
        if (this.logger.isLoggable(Level.INFO)) {
            this.logger.info("JBI exchange received - Process DEL operation");
        }
        // Checks the MEP of the incoming message
        if (!exchange.isInOnlyPattern()) {
            throw new MessagingException(SFTPUtil.getValidMepMessageForOperation(SFTPConstants.DEL,
                    Message.MEPConstants.IN_ONLY_PATTERN));
        }
        final Node root = inDocument.getFirstChild();
        if (root == null || !SFTPConstants.DEL.equalsIgnoreCase(root.getLocalName())) {
            throw new ConfigurationException("Missing root element : " + SFTPConstants.DEL);
        }
        // retrieve the filename to get
        final Node filePatternNode = XMLHelper.findChild(root, null, SFTPConstants.SFTP_FILENAME,
                true);
        String filePattern = null;
        if (filePatternNode == null) {
            throw new ConfigurationException("Missing element : " + SFTPConstants.SFTP_FILENAME);
        } else {
            filePattern = filePatternNode.getTextContent();
        }
        try {
            FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();          
            String flowInstanceId = null;
            String flowStepId = null;
            if (flowAttributes != null) {
                flowInstanceId = flowAttributes.getFlowInstanceId();
                flowStepId = flowAttributes.getFlowStepId();
            }
            
            SftpProvideExtFlowStepBeginLogData sftpProvideExtFlowStepBeginLogData = new SftpProvideExtFlowStepBeginLogData(
                    flowInstanceId, flowStepId, filePattern, client.getInfo().getDirectory());
            this.logger.log(Level.MONIT, "", sftpProvideExtFlowStepBeginLogData);
            
            client.del(filePattern);
        } catch (final NullPointerException ex) {
            final String error = "Can not delete file [" + filePattern + "] on SFTP server ["
                    + client.getInfo().getServer() + "]. Cause : Socket timeout occurs";
            throw new MessagingException(error);
        } catch (final IOException ex) {
            final String error = "Can not delete file [" + filePattern + "] on SFTP server ["
                    + client.getInfo().getServer() + "]. IOException : " + ex.getMessage();
            throw new MessagingException(error);
        }
    }
}
